/* gweather.h
 *
 * SPDX-FileCopyrightText: 2008, Red Hat, Inc.
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#define IN_GWEATHER_H

#include <libgweather/gweather-version.h>
#include <libgweather/gweather-enums.h>
#include <libgweather/gweather-enum-types.h>
#include <libgweather/gweather-info.h>
#include <libgweather/gweather-location.h>

#undef IN_GWEATHER_H
